/*
 * The contents of this file are subject to the AOLserver Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * http://aolserver.lcs.mit.edu/.
 *
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
 * the License for the specific language governing rights and limitations
 * under the License.
 *
 * The Original Code is AOLserver Code and related documentation
 * distributed by AOL.
 * 
 * The Initial Developer of the Original Code is America Online,
 * Inc. Portions created by AOL are Copyright (C) 1999 America Online,
 * Inc. All Rights Reserved.
 *
 * Alternatively, the contents of this file may be used under the terms
 * of the GNU General Public License (the "GPL"), in which case the
 * provisions of GPL are applicable instead of those above.	If you wish
 * to allow use of your version of this file only under the terms of the
 * GPL and not to allow others to use your version of this file under the
 * License, indicate your decision by deleting the provisions above and
 * replace them with the notice and other provisions required by the GPL.
 * If you do not delete the provisions above, a recipient may use your
 * version of this file under either the License or the GPL.
 */

/* 
 * adpAS21.c --
 *
 *	ADP Module for AOLserver 2.1.
 *	
 *	Based on adp.c in AOLserver 3.0 beta 3.	
 *
 */

#include "ns.h"
#include "adpAS21.h"


/*
 * The following structure maintains an ADP call frame.	PushFrame
 * is used to save the previous state of the per-thread AdpData
 * structure in a Frame allocated on the stack and PopFrame restores
 * the previous state from the Frame.
 */

typedef struct {
	int 			   argc;
	char			 **argv;
	char			  *cwd;
	int 			   length;
	Ns_DString		   cwdBuf;
} Frame;

/*
 * The following structure is used as a unique key for compiled pages
 * in the cache.
 */

typedef struct Key {
	dev_t dev;
	ino_t ino;
} Key;

/*
 * Local functions defined in this file.
 */

static void FreePage(Page *pagePtr);
static void Parse(char *filename, Ns_DString *out, char *pagein);
static Tcl_CmdProc AdpProcCmd;
static void (DelAdpData)(void *);
static void ParsePage(Ns_DString *, char *page);
static void PushFrame(Frame *framePtr, char *file, int argc, char **argv);
static void PopFrame(Frame *framePtr);
static int DebugInit(Tcl_Interp *interp, char *host, char *port,
					 char *procs);
static Ns_TclInterpInitProc EnableCmds;
static void TextChunk(Ns_DString *dsPtr, char *text);

/*
 * Static global variables
 */

static int						 expireParm;
static int						 debugParm;
static char 					*defaultParserParm;
static char 					*debugInitProc;
static Ns_ThreadLocalStorage	 adKey;
static char 					*startPage;
static Ns_AdpParserProc			*defParserProc = NULL;
static Tcl_HashTable			 extensionsTable;
static Tcl_HashTable			 parsersTable;
static Ns_Cache *nsAdpCachePtr = NULL;
static char						 nsServer[50];

/*
 * In AS3.0 these are in tclcmds.c NsTclCreateCmds
 */
typedef struct TclCmd {
    char           *name;
    Tcl_CmdProc    *proc;
    ClientData      clientData;
} TclCmd;
static TclCmd AdpTclCmds[] = {
	{
		"ns_puts", NsTclPutsCmd, NULL
	},
	{
		"ns_adp_puts", NsTclPutsCmd, NULL
	},
	{
		"ns_adp_dir", NsTclDirCmd, NULL
	},
	{
		"ns_adp_break", NsTclBreakCmd, (ClientData) ADP_BREAK
	},
	{
		"ns_adp_return", NsTclBreakCmd, (ClientData) ADP_RETURN
	},
	{
		"ns_adp_abort", NsTclBreakCmd, (ClientData) ADP_ABORT
	},
	{
		"ns_adp_exception", NsTclExceptionCmd, NULL
	},
	{
		"ns_adp_argc", NsTclArgcCmd, NULL
	},
	{
		"ns_adp_argv", NsTclArgvCmd, NULL
	},
	{
		"ns_adp_bind_args", NsTclBindCmd, NULL
	},
	{
		"ns_adp_tell", NsTclTellCmd, NULL
	},
	{
		"ns_adp_trunc", NsTclTruncCmd, NULL
	},
	{
		"ns_adp_dump", NsTclDumpCmd, NULL
	},
	{
		"ns_adp_eval", NsTclAdpEvalCmd, NULL
	},
	{
		"ns_adp_stream", NsTclStreamCmd, NULL
	},
	{
		"ns_cache_flush", NsTclFlushCmd, NULL
	},
	{
		"ns_cache_stats", NsTclStatsCmd, NULL
	},
	{
		"ns_cache_names", NsTclNamesCmd, NULL
	},
	{
		"ns_cache_size", NsTclSizeCmd, NULL
	},
	{
		"ns_register_adptag", NsTclRegisterTagCmd, NULL
	},
	{
		"ns_adp_registeradp", NsTclRegisterAdpCmd, NULL
	},
	{
		NULL, NULL, NULL
	}
};


/*
 * Global variables within this source directory
 */
extern char				nsServer[];

/*
 * Non-Static Global Variables
 */
DllExport int Ns_ModuleVersion = 1;

DllExport int
Ns_ModuleInit(char *server, char *hModule) 
{

	char *path;
	int parm;
	
	NsCacheInit(server);

	strcpy(nsServer, server);

	Ns_AllocThreadLocalStorage(&adKey, DelAdpData);

	path = Ns_ConfigPath(server, NULL, "adp", NULL);
	if (Ns_ConfigGetBool(path, "cache", &parm) == NS_FALSE) {
		parm = NS_TRUE;
	}
	if (parm == NS_TRUE) {
		/*
	 	 * Enable page caching.
		 */
		if (Ns_ConfigGetInt(path, "cachesize", &parm) == NS_FALSE ||
			parm < 1) {
			
			parm = 5000 * 1024; /* 5meg */
		}
		
		nsAdpCachePtr = Ns_CacheCreateSz("adp", sizeof(Key) / sizeof(int),
			parm, (Ns_Callback *) FreePage);
	}
	
	/*
	 * Enable various optional features.
	 */
	
	defaultParserParm = Ns_ConfigGet(path, "DefaultParser");
	if (defaultParserParm == NULL) {
		defaultParserParm = "aspStyle";
	}
	if (Ns_ConfigGetBool(path, "enableexpire", &expireParm) == NS_FALSE) {
		expireParm = 0;
	}
	if (Ns_ConfigGetBool(path, "enabledebug", &debugParm) == NS_FALSE) {
		debugParm = 0;
	}
	debugInitProc = Ns_ConfigGet(path, "debuginit");
	if (debugInitProc == NULL) {
		debugInitProc = "ns_adp_debuginit";
	}
	startPage = Ns_ConfigGet(path, "startpage");
	if (startPage != NULL &&
		access(startPage, R_OK) != 0) {
		
		Ns_Log(Error, "ADP: can't read start page %s: %s",
			startPage, strerror(errno));
		startPage = NULL;
	}
	
	/*
  	 * Add the Tcl commands and register ADP proccessor for
	 * requested URLs.
	 */
	Ns_TclInitInterps(server, EnableCmds, NULL);

	/*
	 * Initialize the hash table of parsers
	 */
	Tcl_InitHashTable(&parsersTable, TCL_STRING_KEYS);
	Tcl_InitHashTable(&extensionsTable, TCL_STRING_KEYS);


	/*
	 * Call Parser initalization. In AS3 this is called in main function
	 * in main.c
	 */
	NsAdpParsers();
	
	return NS_OK;
}



/*
 *==========================================================================
 * API functions
 *==========================================================================
 */

/*
 *----------------------------------------------------------------------
 *
 * Ns_AdpRegisterParser --
 *
 *	Sets the ADP parser.
 *	
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	None.
 *
 *----------------------------------------------------------------------
 */

int
Ns_AdpRegisterParser(char *extension, Ns_AdpParserProc *newParserProc)
{
	Tcl_HashEntry *hePtr;
	int 		   new;
	
	if (Ns_InfoServersStarted() == NS_TRUE) {
		Ns_Log(Error, "ADP: attempt to register ADP parser after server startup.");
		return NS_ERROR;
	}
	hePtr = Tcl_CreateHashEntry(&parsersTable, extension, &new);
	Tcl_SetHashValue(hePtr, (void *) newParserProc);
	
	return NS_OK;
}

/* 
 *=========================================================================
 * Exported functions
 *=========================================================================
 */

/*
 *----------------------------------------------------------------------
 *
 * NsAdpParsers --
 *
 *	Load ADP parser mappings. 
 *
 * Results: 
 *	None. 
 *
 * Side effects:
 *	Registers requests for each parser. 
 *
 *----------------------------------------------------------------------
 */

void
NsAdpParsers(void)
{
	char			 *path;
	Tcl_HashEntry	 *hePtr;
	int 			  size = 0;
	Ns_Set			 *setPtr;
	int 			  i, new;
	Ns_AdpParserProc *parserProc;
	
	path = Ns_ConfigPath(nsServer, NULL, "adp", "parsers", NULL);
	
	NsAdpFancyInit(nsServer, path);
	NsAspStyleAdpsInit(nsServer, path);

	Ns_AdpRegisterParser("adp", ParsePage);
	
	if (path != NULL) {
		setPtr = Ns_ConfigGetSection(path);
		size = Ns_SetSize(setPtr);
	}
	
	hePtr = Tcl_FindHashEntry(&parsersTable, defaultParserParm);
	parserProc = (Ns_AdpParserProc *) Tcl_GetHashValue(hePtr);
	defParserProc = parserProc;
	
	hePtr = Tcl_CreateHashEntry(&extensionsTable, ".adp", &new);
	Tcl_SetHashValue(hePtr, defParserProc);
	
	for (i=0; i < size; i++) {
		char			 *parser;
		char			 *ext;
		
		parser = Ns_SetKey(setPtr, i);
		ext = Ns_SetValue(setPtr, i);
		
		hePtr = Tcl_FindHashEntry(&parsersTable, parser);
		if (hePtr == NULL) {
			Ns_Log(Notice, "ADP: invalid parser '%s'", parser);
			continue;
		}
		
		parserProc = (Ns_AdpParserProc *) Tcl_GetHashValue(hePtr);
		hePtr = Tcl_CreateHashEntry(&extensionsTable, ext, &new);
		Tcl_SetHashValue(hePtr, parserProc);
	}
}


/*
 *----------------------------------------------------------------------
 *
 * NsAdpGetData --
 *
 *	Return the per-thread ADP context structure.
 *
 * Results:
 *	Pointer to AdpData.
 *
 * Side effects:
 *	None.
 *
 *----------------------------------------------------------------------
 */

AdpData *
NsAdpGetData(void)
{
	AdpData *adPtr;
	
	Ns_GetThreadLocalStorage(&adKey, &adPtr);
	if (adPtr == NULL) {
		adPtr = ns_calloc(1, sizeof(AdpData));
		Ns_DStringInit(&adPtr->output);
		Ns_SetThreadLocalStorage(&adKey, adPtr);
	}
	
	return adPtr;
}


/*
 *----------------------------------------------------------------------
 *
 * NsAdpCopyChunks --
 *
 *	Copy text/script chunks from an ADP Page structure.
 *
 * Results:
 *	Pointer to per-thread chunk copy which can be executed
 *	with NsAdpEval.
 *
 * Side effects: 
 *	None.
 *
 *----------------------------------------------------------------------  
 */

char *
NsAdpCopyChunks(Page *pagePtr)
{
	char *chunks;
	
	chunks = ns_malloc(pagePtr->length);
	memcpy(chunks, pagePtr->chunks, pagePtr->length);
	return chunks;
}


/*
 *----------------------------------------------------------------------
 *
 * NsAdpFreeChunks --
 *
 *	Frees chunks created with NsAdpCopyChunks.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	None.
 *
 *----------------------------------------------------------------------
 */

void
NsAdpFreeChunks(char *chunks)
{
	ns_free(chunks);
}


/*
 *----------------------------------------------------------------------
 *
 * NsAdpEval --
 *
 *	Evaluate a list of chunks from an ADP.
 *
 * Results:
 *	TCL_OK or TCL_ERROR.
 *
 * Side effects:
 *	Depends on script chunks.
 *
 *----------------------------------------------------------------------
 */

int
NsAdpEval(Tcl_Interp *interp, char *file, char *chunks)
{
	int chunk, n, code, fd;
	register char *ch = chunks;
	Ns_DString ds;
	char buf[10], *script, debugfile[255];
	AdpData *adPtr;
	
	adPtr = NsAdpGetData();

// Ns_CheckStatck does not exist in AS 2.1.
//	if (Ns_CheckStack() != NS_OK) {
//		interp->result =  "danger: stack grown too large (recursive adp?)";
//		adPtr->exception = ADP_OVERFLOW;
//		return TCL_ERROR;
//	}
	
	if (file == NULL) {
		file = "<inlined script>";
	}
	Ns_DStringInit(&ds);
	code = TCL_OK;
	chunk = 1;
	
	while (*ch && adPtr->exception == ADP_OK) {
		n = strlen(ch);
		if (*ch++ == 't') {
			Ns_DStringNAppend(&adPtr->output, ch, n-1);
		} else {
			script = ch;
			if (adPtr->debugLevel > 0) {
				Ns_DStringTrunc(&ds, 0);
				sprintf(buf, "%d", adPtr->debugLevel);
				Ns_DStringVarAppend(&ds,
					"#\n"
					"# level: ", buf, "\n", NULL);
				sprintf(buf, "%d", chunk);
				Ns_DStringVarAppend(&ds,
					"# chunk: ", buf, "\n"
					"# file:  ", file, "\n"
					"#\n\n", ch, NULL);
				sprintf(debugfile, "/tmp/adp%d.%d.XXXXXX",
					adPtr->debugLevel, chunk);
				mktemp(debugfile);
				fd = open(debugfile, O_WRONLY|O_TRUNC|O_CREAT, 0644);
				if (fd < 0) {
					Ns_Log(Notice, "ADP: could not open %s:  %s", debugfile,
						strerror(errno));
				} else {
					write(fd, ds.string, ds.length);
					close(fd);
					Ns_DStringTrunc(&ds, 0);
					Ns_DStringVarAppend(&ds, "source ", debugfile, NULL);
					script = ds.string;
				}
			}
			
			code = Tcl_Eval(interp, script);
			if (code != TCL_OK &&
				code != TCL_RETURN &&
				adPtr->exception == ADP_OK) {
				
				Ns_DStringTrunc(&ds, 0);
				sprintf(buf, "%d", chunk);
				Ns_DStringAppend(&ds, "\n    invoked in chunk: ");
				Ns_DStringAppend(&ds, buf);
				Ns_DStringAppend(&ds, " of adp: ");
				Ns_DStringAppend(&ds, file);
				Tcl_AddErrorInfo(interp, ds.string);
				Ns_TclLogError(interp);
			}
			if (code == TCL_ERROR && adPtr->exception == ADP_RETURN) {
				adPtr->exception = ADP_OK;
				code = TCL_OK;
			}
			if (script != ch) {
				unlink(debugfile);
			}
			++chunk;
		}
		ch += n;
		NsAdpFlush(adPtr);
	}
	NsAdpFlush(adPtr);
	Ns_DStringFree(&ds);
	
	return code;
}


/*
 *----------------------------------------------------------------------
 *
 * NsAdpStreamOn --
 *
 *	Turn on streaming for the current connection.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	None unless streaming is not already enabled in which case
 *	headers and any pending content is immediately sent to the
 *	client.  Note that no content-length header will be sent
 *	which has the side effect of disabling connection: keep-alive
 *	for the client.
 *
 *----------------------------------------------------------------------
 */

void
NsAdpStreamOn(void)
{
	AdpData *adPtr;
	
	adPtr = NsAdpGetData();
	if (adPtr->conn == NULL) {
		return;
	}
	if (adPtr->fStream != NS_TRUE) {
		adPtr->fStream = NS_TRUE;
		Ns_ConnSetRequiredHeaders(adPtr->conn, "text/html", 0);
		Ns_ConnFlushHeaders(adPtr->conn, 200);
	}
	NsAdpFlush(adPtr);
}


/*
 *----------------------------------------------------------------------
 *
 * NsAdpFlush --
 *
 *	Flush current output.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	None unless streaming is enabled in which case content is
 *	written directly out the connection.  Also, output is disabled
 *	during inlined evalution where the output buffer is used to
 *	accumulate the inlined evaluation result before being copied
 *	to the interp.
 *
 *----------------------------------------------------------------------
 */

void
NsAdpFlush(AdpData *adPtr)
{
	if (adPtr->fStream == NS_TRUE &&
		adPtr->evalLevel == 0 &&
		adPtr->output.length > 0) {
		
		Ns_WriteConn(adPtr->conn, adPtr->output.string,
			adPtr->output.length);
		Ns_DStringTrunc(&adPtr->output, 0);
	}
}


/*
 *----------------------------------------------------------------------
 *
 * NsTclAdpEvalCmd --
 *
 *	Process the Tcl _ns_adp_eval command.
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	Page string is parsed and evaluated at current Tcl level in a
 *	new ADP call frame.
 *
 *----------------------------------------------------------------------
 */

int
NsTclAdpEvalCmd(ClientData dummy, Tcl_Interp *interp, int argc,
				char **argv)
{
	AdpData 		 *adPtr;
	Frame			  frame;
	Ns_DString		  ds;
	int 			  code;
	int 			  offset = 0;
	int 			  parser = -1;
	Tcl_HashEntry	 *hePtr;
	Ns_AdpParserProc *parserProc;
	
	if (argc >= 2) {
		if (!strcmp(argv[1], "-parser")) {
			offset += 2;
			parser = 1;
		}
	}
	if (argc < 2 + offset) {
		Tcl_AppendResult(interp, "wrong # args: should be \"",
			argv[0], " ?-parser parser? page ?arg ...?\"", NULL);
		return TCL_ERROR;
	}
	
	if (parser != -1) {
		hePtr = Tcl_FindHashEntry(&parsersTable, argv[parser]);
		if (hePtr == NULL) {
			Tcl_AppendResult(interp, "invalid parser \"", parser, "\"", NULL);
			return TCL_ERROR;
		}
		parserProc = (Ns_AdpParserProc *) Tcl_GetHashValue(hePtr);
	} else {
		parserProc = defParserProc;
	}
	
	/*
	 * Increment the inline eval level to ensure flushing is disabled,
	 * push a frame, execute the code, and then more any result to the
	 * interp from the output buffer.
	 */
	
	Ns_DStringInit(&ds);
	
	adPtr = NsAdpGetData();
	++adPtr->evalLevel;
	PushFrame(&frame, NULL, argc-1-offset, argv+1+offset);
	Parse(NULL, &ds, argv[1]);
	code = NsAdpEval(interp, argv[0], ds.string);
	if (adPtr->output.length > frame.length) {
		Tcl_SetResult(interp, adPtr->output.string + frame.length,
			TCL_VOLATILE);
		Ns_DStringTrunc(&adPtr->output, frame.length);
	}
	PopFrame(&frame);
	--adPtr->evalLevel;
	
	Ns_DStringFree(&ds);
	
	return code;
}


/*
 *----------------------------------------------------------------------
 *
 * NsTclIncludeCmd --
 *
 *	Process the Tcl _ns_adp_include and _ns_adp_parse commands.
 *	This routines is the core ADP execution engine.
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	Parse page for the given file is fetched from the cache and
 *	evaluated using the Tcl76 or Tcl81 code engine.
 *
 *----------------------------------------------------------------------
 */

int
NsTclIncludeCmd(ClientData parse, Tcl_Interp *interp, int argc,
				char **argv)
{
	struct stat st;
	Ns_DString file, ds;
	void *codePtr;
	AdpData *adPtr;
	Frame frame;
	Key key;
	Page *pagePtr;
	Ns_Entry *ePtr;
	int code, fd, n, new;
	char *p;
	
	if (argc < 2) {
		Tcl_AppendResult(interp, "wrong # args: should be \"",
			argv[0], " file ?args ...?\"", NULL);
		return TCL_ERROR;
	}
	
	Ns_DStringInit(&ds);
	Ns_DStringInit(&file);
	adPtr = NsAdpGetData();
	codePtr = NULL; 	
	pagePtr = NULL;
	code = TCL_ERROR;
	
	/*
	 * Construct the full, normalized path to the ADP file.
	 */

	/* 
	 * Modified for AS21: The original expression was argv[1][0] == '/'. 
	 * The purpose of this is to detect if we were based ab absolute 
	 * or relative path.
	 */
	if (argv[1][1] == ':') {
		Ns_NormalizePath(&file, argv[1]);
	} else {
		Ns_MakePath(&ds, adPtr->cwd, argv[1], NULL);
		Ns_NormalizePath(&file, ds.string);
		Ns_DStringTrunc(&ds, 0);
	}

	
	/*
	 * Check for TclPro debugging.
	 */
	
	if (adPtr->debugLevel > 0) {
		++adPtr->debugLevel;
	} else if (debugParm != NS_FALSE &&
		adPtr->debugFile != NULL &&
		(p = strrchr(file.string, '/')) != NULL &&
		Tcl_StringMatch(p+1, adPtr->debugFile)) {
		
		Ns_Conn *conn;
		Ns_Set *hdrs;
		char *host, *port, *procs;
		
		conn = Ns_TclGetConn(interp);
		hdrs = Ns_ConnGetQuery(conn);
		host = Ns_SetIGet(hdrs, "dhost");
		port = Ns_SetIGet(hdrs, "dport");
		procs = Ns_SetIGet(hdrs, "dprocs");
		if (DebugInit(interp, host, port, procs) != TCL_OK) {
			Ns_ConnReturnNotice(conn, 200, "Debug Init Failed",
				interp->result);
			adPtr->exception = ADP_ABORT;
			goto done;
		}
	}
	
	/*
 	 * Verify the file is an existing, ordinary file.
	 */
	
	if (stat(file.string, &st) != 0) {
		Tcl_AppendResult(interp, "could not stat \"",
			file.string, "\": ", Tcl_PosixError(interp), NULL);
	} else if (S_ISREG(st.st_mode) == 0) {
		Tcl_AppendResult(interp, "not an ordinary file: ", file.string,
			NULL);
	} else {
		key.dev = st.st_dev;
		key.ino = st.st_ino;
		pagePtr = NULL;
		
		/*
		 * If caching is enabled, search and validate an existing cache
		 * entry for the page.
		 */
 
		if (nsAdpCachePtr != NULL) {
			Ns_CacheLock(nsAdpCachePtr);
			ePtr = Ns_CacheCreateEntry(nsAdpCachePtr, (char *) &key, &new);
			if (new == 0) {
				pagePtr = Ns_CacheGetValue(ePtr);
				if (pagePtr->mtime != (time_t) st.st_mtime 
					|| pagePtr->size != (size_t) st.st_size) {
					Ns_CacheUnsetValue(ePtr);
					pagePtr = NULL;
				}
			}
		}

		
		/*
		 * Read and compile the page if it was not found in the cache.
		 */
		
		if (pagePtr == NULL) {
			fd = open(file.string, O_RDONLY);
			if (fd < 0) {
				Tcl_AppendResult(interp, "could not open \"",
					file, "\": ", Tcl_PosixError(interp), NULL);
			} else {
				p = ns_malloc(st.st_size + 1);
				n = read(fd, p, st.st_size);
				close(fd);
				if (n < 0) {
					Tcl_AppendResult(interp, "read() failed: ",
						Tcl_PosixError(interp), NULL);
				} else {
					p[n] = '\0';
					Parse(file.string, &ds, p);
					pagePtr = ns_malloc(sizeof(Page) + ds.length);
					pagePtr->mtime = st.st_mtime;
					pagePtr->size = st.st_size;
					pagePtr->length = ds.length+1;
					pagePtr->pdPtr = NULL;
					memcpy(pagePtr->chunks, ds.string, ds.length+1);
					Ns_DStringTrunc(&ds, 0);
				}
				ns_free(p);
			}
			
			/*
			 * If caching is enabled, entry the compiled page in the
			 * cache or delete the cache entry if the compile failed.
			 */
			if (nsAdpCachePtr != NULL) {
				if (pagePtr != NULL) {
					Ns_CacheSetValueSz(ePtr, pagePtr, pagePtr->size);
				} else {				
					Ns_CacheDeleteEntry(ePtr);
				}
			}
		}
		
		/*
		 * If a page was found, get a per-thread code object which can
		 * be safely executed below.  If caching is disabled, the
		 * per-thread object is simply page chunks as the new page is
		 * private to this thread.  Otherwise the code is either a copy
		 * of the page chunks for raw ADP debugging evaluation or some
		 * Tcl version specific object.
		 */
		
		if (pagePtr != NULL) {
			if (nsAdpCachePtr == NULL) {
				codePtr = pagePtr->chunks;
			} else {
				codePtr = (void *) NsAdpCopyChunks(pagePtr);
				Ns_CacheUnlock(nsAdpCachePtr);
			}
		}
	}
	
	/*
	 * If a valid code object was obtained, push a new call frame, run and
	 * release the code, and pop the frame.	When caching is disabled
	 * or in debug mode the code is executed as chunks by NsAdpEval.
  	 * Otherwise the code is executed by the Tcl version specific
	 * evaluation engine.
	 */
	
	if (codePtr != NULL) {
		PushFrame(&frame, file.string, argc-1, argv+1);
		if (adPtr->debugLevel > 0 || nsAdpCachePtr == NULL) {
			code = NsAdpEval(interp, file.string, (char *) codePtr);
			if (nsAdpCachePtr != NULL) {
				NsAdpFreeChunks((char *) codePtr);
			}
		} else {
			code = NsAdpEval(interp, file.string, codePtr);
			NsAdpFreeChunks(codePtr);
		}
		if (parse && code == TCL_OK &&
			adPtr->output.length > frame.length) {
			
			Tcl_SetResult(interp, adPtr->output.string + frame.length,
				TCL_VOLATILE);
			Ns_DStringTrunc(&adPtr->output, frame.length);
		}
		PopFrame(&frame);
		NsAdpFlush(adPtr);
	}
	
	/*
	 * If caching is disabled, free the per-thread compiled page.
	 */
	
	if (pagePtr != NULL && nsAdpCachePtr == NULL) {
		ns_free(pagePtr);
	}
	
	if (adPtr->debugLevel > 0) {
		--adPtr->debugLevel;
	}
	
done:
	Ns_DStringFree(&file);
	Ns_DStringFree(&ds);
	
	return code;
}


/*
 *----------------------------------------------------------------------
 *
 * NsTclPutsCmd --
 *
 *	Process the Tcl ns_adp_puts command to append output.
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	Output buffer is extended with given text.
 *
 *----------------------------------------------------------------------
 */

int
NsTclPutsCmd(ClientData ignored, Tcl_Interp *interp, int argc, char **argv)
{
	AdpData *adPtr;
	
	if (argc != 2 && argc != 3) {
		Tcl_AppendResult(interp, "wrong # args: should be \"",
			argv[0], " ?-nonewline? string\"", NULL);
		return TCL_ERROR;
	}
	if (argc == 3 &&
		strcmp(argv[1], "-nonewline") != 0) {
		
		Tcl_AppendResult(interp, "invalid flag \"",
			argv[1], "\": expected -nonewline", NULL);
		return TCL_ERROR;
	}
	
	adPtr = NsAdpGetData();
	Ns_DStringAppend(&adPtr->output, argv[argc-1]);
	if (argc != 3) {
		Ns_DStringNAppend(&adPtr->output, "\n", 1);
	}
	NsAdpFlush(adPtr);
	
	return TCL_OK;
}



/*
 *----------------------------------------------------------------------
 *
 * NsTclDirCmd --
 *
 *	Process the Tcl ns_adp_dir command to return the current ADP
 *	directory.
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	None.
 *
 *----------------------------------------------------------------------
 */

int
NsTclDirCmd(ClientData ignored, Tcl_Interp *interp, int argc, char **argv)
{
	AdpData *adPtr;
	
	if (argc != 1) {
		Tcl_AppendResult(interp, "wrong # args: should be \"",
			argv[0], "\"", NULL);
		return TCL_ERROR;
	}
	adPtr = NsAdpGetData();
	if (adPtr->cwd != NULL && *adPtr->cwd) {   
		Tcl_SetResult(interp, adPtr->cwd, TCL_VOLATILE);
	} else {
		interp->result = "/";
	}
	
	return TCL_OK;
}


/*
 *----------------------------------------------------------------------
 *
 * NsTclBreakCmd --
 *
 *	Process the Tcl ns_adp_break and ns_adp_abort commands to halt
 *	page generation.
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	Break or abort exception is noted and will be handled in
 *	AdpProc.
 *
 *----------------------------------------------------------------------
 */

int
NsTclBreakCmd(ClientData clientData, Tcl_Interp *interp, int argc,
			  char **argv)
{
	AdpData *adPtr;
	
	if (argc != 1 && argc != 2) {
		Tcl_AppendResult(interp, "wrong # args: should be \"",
			argv[0], " ?retval?\"", NULL);
		return TCL_ERROR;
	}
	adPtr = NsAdpGetData();
	adPtr->exception = (int) clientData;
	if (argc == 2) {
		Tcl_AppendResult(interp, argv[1], NULL);
	}
	return TCL_ERROR;
}


/*
 *----------------------------------------------------------------------
 *
 * NsTclTellCmd --
 *
 *	Process the Tcl ns_adp_tell commands to return the current
 *	offset within the output buffer.
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	None.
 *
 *----------------------------------------------------------------------
 */

int
NsTclTellCmd(ClientData ignored, Tcl_Interp *interp, int argc, char **argv)
{
	AdpData *adPtr;
	
	if (argc != 1) {
		Tcl_AppendResult(interp, "wrong # args: should be \"",
			argv[0], "\"", NULL);
		return TCL_ERROR;
	}
	adPtr = NsAdpGetData();
	sprintf(interp->result, "%d", adPtr->output.length);
	
	return TCL_OK;
}


/*
 *----------------------------------------------------------------------
 *
 * NsTclTruncCmd --
 *
 *	Process the Tcl ns_adp_trunc commands to truncate the output
 *	buffer to the given length.
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	Output buffer is truncated.
 *
 *----------------------------------------------------------------------
 */

int
NsTclTruncCmd(ClientData ignored, Tcl_Interp *interp, int argc,
			  char **argv)
{
	AdpData *adPtr;
	int length;
	
	if (argc != 1 && argc != 2) {
		Tcl_AppendResult(interp, "wrong # args: should be \"",
			argv[0], " ?length?\"", NULL);
		return TCL_ERROR;
	}
	if (argc == 1) {
		length = 0;
	} else {
		if (Tcl_GetInt(interp, argv[1], &length) != TCL_OK) {
			return TCL_ERROR;
		}
		if (length < 0) {
			Tcl_AppendResult(interp, "invalid length: ", argv[1], NULL);
			return TCL_ERROR;
		}
	}
	adPtr = NsAdpGetData();
	Ns_DStringTrunc(&adPtr->output, length);
	
	return TCL_OK;
}


/*
 *----------------------------------------------------------------------
 *
 * NsTclDumpCmd --
 *
 *	Process the Tcl ns_adp_dump commands to return the entire text
 *	of the output buffer.
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	None.
 *
 *----------------------------------------------------------------------
 */

int
NsTclDumpCmd(ClientData ignored, Tcl_Interp *interp, int argc, char **argv)
{
	AdpData *adPtr;
	
	if (argc != 1) {
		Tcl_AppendResult(interp, "wrong # args: should be \"",
			argv[0], "\"", NULL);
		return TCL_ERROR;
	}
	adPtr = NsAdpGetData();
	Tcl_SetResult(interp, adPtr->output.string, TCL_VOLATILE);
	
	return TCL_OK;
}


/*
 *----------------------------------------------------------------------
 *
 * NsTclArgcCmd --
 *
 *	Process the Tcl ns_adp_args commands to return the number of
 *	arguments in the current ADP frame.
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	None.
 *
 *----------------------------------------------------------------------
 */

int
NsTclArgcCmd(ClientData ignored, Tcl_Interp *interp, int argc, char **argv)
{
	AdpData *adPtr;
	
	if (argc != 1) {
		Tcl_AppendResult(interp, "wrong # args: should be \"",
			argv[0], "\"", NULL);
		return TCL_ERROR;
	}
	adPtr = NsAdpGetData();
	sprintf(interp->result, "%d", adPtr->argc);
	
	return TCL_OK;
}


/*
 *----------------------------------------------------------------------
 *
 * NsTclArgvCmd --
 *
 *	Process the Tcl ns_adp_args commands to return an argument (or
 *	the entire list of arguments) within the current ADP frame.
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	None.
 *
 *----------------------------------------------------------------------
 */

int
NsTclArgvCmd(ClientData ignored, Tcl_Interp *interp, int argc, char **argv)
{
	AdpData *adPtr;
	int i;
	
	if (argc != 1 && argc != 2) {
		Tcl_AppendResult(interp, "wrong # args: should be \"",
			argv[0], " ?index?\"", NULL);
		return TCL_ERROR;
	}
	adPtr = NsAdpGetData();
	if (adPtr->argv != NULL) {
		if (argc == 1) {
			for (i = 0; i < adPtr->argc; ++i) {
				Tcl_AppendElement(interp, adPtr->argv[i]);
			}
		} else {
			if (Tcl_GetInt(interp, argv[1], &i) != TCL_OK) {
				return TCL_ERROR;
			}
			if (i > adPtr->argc) {
				i = adPtr->argc;
			}
			Tcl_SetResult(interp, adPtr->argv[i], TCL_VOLATILE);
		}
	}
	
	return TCL_OK;
}


/*
 *----------------------------------------------------------------------
 *
 * NsTclBindCmd --
 *
 *	Process the Tcl ns_adp_bind_args commands to copy arguements
 *	from the current frame into local variables.
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	One or more local variables are created.
 *
 *----------------------------------------------------------------------
 */

int
NsTclBindCmd(ClientData ignored, Tcl_Interp *interp, int argc, char **argv)
{
	AdpData *adPtr;
	int i;
	char *arg;
	
	if (argc < 2) {
		Tcl_AppendResult(interp, "wrong # args: should be \"",
			argv[0], " varName ?varName ...?\"", NULL);
		return TCL_ERROR;
	}
	adPtr = NsAdpGetData();
	if (adPtr->argc == 0) {
		Tcl_AppendResult(interp, "not in an ADP", NULL);
		return TCL_ERROR;
	}
	if (adPtr->argc != argc) {
		char buf[sizeof(int) * 3 + 1];
		
		sprintf(buf, "%d", adPtr->argc - 1);
		Tcl_AppendResult(interp, "wrong # args: this ADP was passed ",
			buf, " parameters", NULL);
		return TCL_ERROR;
	}
	
	for (i = 1; i < argc; ++i) {
		if (adPtr->argv != NULL && i < adPtr->argc) {
			arg = adPtr->argv[i];
		} else {
			arg = "";
		}
		if (Tcl_SetVar(interp, argv[i], arg, TCL_LEAVE_ERR_MSG) == NULL) {
			return TCL_ERROR;
		}
	}
	
	return TCL_OK;
}			


/*
 *----------------------------------------------------------------------
 *
 * NsTclExcepetionCmd --
 *
 *	Process the Tcl ns_adp_exception commands to return the current
 *	exception state, ok, abort, or break.
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	None.
 *
 *----------------------------------------------------------------------
 */

int
NsTclExceptionCmd(ClientData ignored, Tcl_Interp *interp, int argc,
				  char **argv)
{
	AdpData *adPtr;
	char *exception;
	
	if (argc != 1 && argc != 2) {
		Tcl_AppendResult(interp, "wrong # args: should be \"",
			argv[0], " ?varName?\"", NULL);
		return TCL_ERROR;
	}
	adPtr = NsAdpGetData();
	if (adPtr->exception == ADP_OK) {
		Tcl_SetResult(interp, "0", TCL_STATIC);
	} else {
		Tcl_SetResult(interp, "1", TCL_STATIC);
	}
	if (argc == 2) {
		switch (adPtr->exception) {
		case ADP_OK:
			exception = "ok";
			break;
		case ADP_BREAK:
			exception = "break";
			break;
		case ADP_ABORT:
			exception = "abort";
			break;
		case ADP_OVERFLOW:
			exception = "overflow";
			break;
		case ADP_RETURN:
			exception = "return";
			break;
		default:
			exception = "unknown";
			break;
		}
		if (Tcl_SetVar(interp, argv[1], exception, 
			TCL_LEAVE_ERR_MSG) == NULL) {
			return TCL_ERROR;
		}
	}
	
	return TCL_OK;
}


/*
 *----------------------------------------------------------------------
 *
 * NsTclStreamCmd --
 *
 *	Process the Tcl ns_adp_stream commands to enable streaming
 *	output mode.
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	See comments for NsAdpStreamOn.
 *
 *----------------------------------------------------------------------
 */

int
NsTclStreamCmd(ClientData ignored, Tcl_Interp *interp, int argc,
			   char **argv)
{
	if (argc != 1) {
		Tcl_AppendResult(interp, "wrong # args: should be \"",
			argv[0], NULL);
		return TCL_ERROR;
	}
	NsAdpStreamOn();
	
	return TCL_OK;
}


/*
 *----------------------------------------------------------------------
 *
 * NsTclDebugCmd --
 *
 *	Process the Tcl ns_adp_debug command to connect to the TclPro
 *	debugger if not already connected.
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	See comments for DebugInit().
 *
 *----------------------------------------------------------------------
 */

int
NsTclDebugCmd(ClientData ignored, Tcl_Interp *interp, int argc,
			  char **argv)
{
	AdpData *adPtr;
	char *host, *port, *procs;
	
	if (argc > 4) {
		Tcl_AppendResult(interp, "wrong # args: should be \"",
			argv[0], " ?procs? ?host? ?port?\"", NULL);
		return TCL_ERROR;
	}
	procs = (argc > 1) ? argv[1] : NULL;
	host = (argc > 2) ? argv[2] : NULL;
	port = (argc > 3) ? argv[3] : NULL;
	
	if (DebugInit(interp, host, port, procs) != TCL_OK) {
		Tcl_SetResult(interp, "could not initialize debugger", TCL_STATIC);
		return TCL_ERROR;
	}
	
	adPtr = NsAdpGetData();
	sprintf(interp->result, "%d", adPtr->debugLevel);
	
	return TCL_OK;
}


/*
 *=========================================================================
 * Static functions
 *=========================================================================
 */

/*
 *----------------------------------------------------------------------
 *
 * EnableCmds --
 *
 *	Enable the ADP Tcl commands and create ns_adp_include and
 *	ns_adp_parse commands.
 *
 * Results:
 *	NS_OK.
 *
 * Side effects:
 *	ADP commands are created in the parent Tcl interp.
 *
 *----------------------------------------------------------------------
 */

static int
EnableCmds(Tcl_Interp *interp, void *ignored)
{
	struct TclCmd *cmd;
	char *path, *map, nsRegisterWrapper[100];
	Ns_Set *set;
	int i;

	char incWrapper[] = 
		"proc ns_adp_include args {\n\teval _ns_adp_include $args\n}";
	char parseWrapper[] =
		"proc ns_adp_parse args {\n\teval _ns_adp_parse $args\n}";
	char as21TclAdpWrapper[] =
			"proc as21TclAdpWrapper {conn ignore} {\n\t_ns_adp_process $conn [ns_conn url $conn]\n}";

	
	/*
	 * Modified for AS21: Create Tcl wrapper hack to get around not being able to 
	 * iniatlize a Tcl interpreter in 2.1
	 */
	Tcl_GlobalEval(interp, as21TclAdpWrapper);

	/*
	 * Modified for AS21: Register ADP for any requested URLs. This is normally 
	 * in NsAdpInit (or in our case Ns_ModuleInit) put since we can't use 
	 * Ns_RegisterRequest and instead have to use a Tcl procedure, we have to 
	 * place the code here (since this the only place where we have access to a 
	 * Tcl interpreter).
	 */
	
	path = Ns_ConfigPath(nsServer, NULL, "adp", NULL);
	map = NULL;
	set = Ns_ConfigGetSection(path);
	for (i = 0; set != NULL && i < Ns_SetSize(set); ++i) {
		char		  *key;
		
		key = Ns_SetKey(set, i);
		if (!stricmp(key, "map")) {
			map = Ns_SetValue(set, i);

			sprintf(nsRegisterWrapper, "ns_register_proc GET %s as21TclAdpWrapper", map);
			Tcl_GlobalEval(interp, nsRegisterWrapper);
			sprintf(nsRegisterWrapper, "ns_register_proc POST %s as21TclAdpWrapper", map);
			Tcl_GlobalEval(interp, nsRegisterWrapper);
			sprintf(nsRegisterWrapper, "ns_register_proc HEAD %s as21TclAdpWrapper", map);
			Tcl_GlobalEval(interp, nsRegisterWrapper);

			Ns_Log(Notice, "ADP: mapped %s", map);
		}
	}
	if (map == NULL) {
		Ns_Log(Notice, "ADP: no Map configuration - disabled");
	}
	


	Tcl_CreateCommand(interp, "_ns_adp_include",
				NsTclIncludeCmd, (ClientData) 0, NULL);
	Tcl_CreateCommand(interp, "_ns_adp_parse", 
				NsTclIncludeCmd, (ClientData) 1, NULL);
	Tcl_CreateCommand(interp, "_ns_adp_process", AdpProcCmd,
			  (ClientData) 1, NULL);
	
	Tcl_GlobalEval(interp, incWrapper);
	Tcl_GlobalEval(interp, parseWrapper);
	

	/*
	 * In AS3 these are in tclcmds.c NsTclCreateCmds
	 */
    cmd = AdpTclCmds;
    while (cmd->name != NULL) {
		Tcl_CreateCommand(interp, cmd->name, cmd->proc, cmd->clientData,
			  NULL);
        ++cmd;
    }


	return NS_OK;
}


/*
 *----------------------------------------------------------------------
 *
 * AdpProcCmd --
 *
 *	Process a request for an ADP page.
 *
 * Results:
 *	A standard AOLserver request result.
 *
 * Side effects:
 *	Depends on code embedded within page.
 *
 *----------------------------------------------------------------------
 */

static int
AdpProcCmd(ClientData parse, Tcl_Interp *interp, int argc,
				char **realArgv)
{
	Ns_Conn          *conn;
	AdpData 		 *adPtr;
	Ns_DString		  file;
	int 			  status;
	char			 *argv[3];
	Frame			  frame;
	Ns_Set			 *setPtr;


	if (argc < 3) {
		Tcl_AppendResult(interp, "wrong # args: should be \"",
			realArgv[0], " conn url\"", NULL);
		return TCL_ERROR;
	}

	conn = (Ns_Conn *) Ns_TclGetConn(interp);

	/*
	 * Verify the ADP file exists and, if so, push a new
	 * call frame and execute it.
	 */
	status = NS_ERROR;
	Ns_DStringInit(&file);
	Ns_UrlToFile(&file, nsServer, realArgv[2]);

	if (access(file.string, R_OK) != 0) {
		status = Ns_ConnReturnNotFound(conn);
	} else {
		adPtr = NsAdpGetData();
		adPtr->conn = conn;
		adPtr->fStream = NS_FALSE;

		if (debugParm &&
			STREQ(conn->request->method, "GET") &&
			(setPtr = Ns_ConnGetQuery(conn)) != NULL) {
			
			adPtr->debugFile = Ns_SetIGet(setPtr, "debug");
		}

		argv[0] = "_ns_adp_include";
		argv[1] = startPage ? startPage : file.string;
		argv[2] = NULL;
		PushFrame(&frame, file.string, 0, NULL);
		
		/*
		 * Set the old conn variable for backwards compatibility.
		 */
		
		Tcl_SetVar2(interp, "conn", NULL, "adp", TCL_GLOBAL_ONLY);
		Tcl_ResetResult(interp);
		
		/*
		 * Ignore error - will be reported in NsAdpEval.
		 */
		NsTclIncludeCmd(NULL, interp, 2, argv);
		
		switch (adPtr->exception) {
		case ADP_ABORT:
			/*
			 * Abort is normally used after a call to a
			 * ns_return function so no response is sent here.
			 */
			status = NS_OK;
			break;
			
		case ADP_OVERFLOW:
			Ns_Log(Error, "ADP: stack overflow:  %s", file.string);
			status = Ns_ConnReturnInternalError(conn);
			break;
			
		default:
			if (expireParm) {
				Ns_ConnSetHeaders(conn, "Expires", "now");
			}
			if (Ns_ConnResponseStatus(conn) == 0) {
				status = Ns_ConnReturnHtml(conn, 200,
					adPtr->output.string,
					adPtr->output.length);
			}
			break;
		}
		PopFrame(&frame);
		
		/*
		 * Cleanup the per-thead ADP context.
		 */
		
		Ns_DStringTrunc(&adPtr->output, 0);
		adPtr->exception = ADP_OK;
		adPtr->depth = 0;
		adPtr->argc = 0;
		adPtr->argv = NULL;
		adPtr->cwd = NULL;
		adPtr->debugLevel = 0;
		adPtr->debugInit = 0;
		adPtr->debugFile = NULL;
	}
	
	Ns_DStringFree(&file);
	
	return status;
}


/*
 *----------------------------------------------------------------------
 *
 * DelAdpData --
 *
 *	AdpData TLS cleanup procedure.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	AdpData structure is free'ed from per-thread memory pool.
 *
 *----------------------------------------------------------------------
 */

static void
DelAdpData(void *arg)
{
	AdpData *adPtr = arg;
	
	Ns_DStringFree(&adPtr->output);
	ns_free(adPtr);
}


/*
 *----------------------------------------------------------------------
 *
 * ParsePage --
 *
 *	Parse an ADP page string into text/script chunks.
 * 
 * Results:
 *	None.
 *
 * Side effects:
 *	Chunks are parsed into given Ns_DString.
 *
 *----------------------------------------------------------------------
 */

static void
ParsePage(Ns_DString *dsPtr, char *page)
{
	register char *s, *e;
	register char *t = page;
	
	while ((s = strstr(t, "<%")) != NULL &&
		(e = strstr(s, "%>")) != NULL) {
		
		*s = '\0';
		TextChunk(dsPtr, t);
		s += 2;
		Ns_DStringNAppend(dsPtr, "s", 1);
		if (*s == '=') {
			Ns_DStringAppend(dsPtr, "ns_adp_puts -nonewline ");
			++s;
		}
		*e = '\0';
		Ns_DStringNAppend(dsPtr, s, e-s+1);
		t = e+2;
	}
	TextChunk(dsPtr, t);
}

static void
TextChunk(Ns_DString *dsPtr, char *text)
{
	Ns_DStringNAppend(dsPtr, "t", 1);
	Ns_DStringNAppend(dsPtr, text, strlen(text)+1);
}


/*
 *----------------------------------------------------------------------
 *
 * PushFrame --
 *
 *	Push an ADP call frame on the ADP stack.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	The given Frame is initialized, the current working directory
 *	is determined from the absolute filename (if not NULL), the
 *	previous state of the per-thread AdpData structure is saved
 * 	and then updated with the current call's arguments.
 *
 *----------------------------------------------------------------------
 */

static void
PushFrame(Frame *framePtr, char *file, int argc, char **argv)
{
	register AdpData *adPtr;
	register char	 *slash;
	
	/*
	 * Save current AdpData state.
	 */
	
	adPtr = NsAdpGetData();
	framePtr->cwd = adPtr->cwd;
	framePtr->length = adPtr->output.length;
	framePtr->argc = adPtr->argc;
	framePtr->argv = adPtr->argv;
	adPtr->argc = argc;
	adPtr->argv = argv;
	++adPtr->depth;
	
	/*
	 * If file is not NULL it indicates a call from
	 * AdpProc or NsTclIncludeCmd.  If so, update the
	 * current working directory based on the
	 * absolute file pathname.
	 */
	
	Ns_DStringInit(&framePtr->cwdBuf);
	if (file != NULL) {
		slash = strrchr(file, '/');
		Ns_DStringNAppend(&framePtr->cwdBuf, file, slash - file);
		adPtr->cwd = framePtr->cwdBuf.string;
	}
}


/*
 *----------------------------------------------------------------------
 *
 * PopFrame --
 *
 *	Pop a previously pushed ADP call frame from the ADP stack.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	Previous state of the per-thread AdpData structure is restored
 *	and the Frame is free'ed.
 *
 *----------------------------------------------------------------------
 */

static void
PopFrame(Frame *framePtr)
{
	register AdpData *adPtr;
	
	adPtr = NsAdpGetData();
	adPtr->argc = framePtr->argc;
	adPtr->argv = framePtr->argv;
	adPtr->cwd = framePtr->cwd;
	--adPtr->depth;
	Ns_DStringFree(&framePtr->cwdBuf);
}


/*
 *----------------------------------------------------------------------
 *
 * DebugInit --
 *
 *	Initialize the debugger by calling the debug init proc with
 *	the hostname and port of the debugger and a pattern of procs
 *	to auto-instrument.
 *
 * Results:
 *	TCL_OK if debugger initialized, TCL_ERROR otherwise.
 *
 * Side effects:
 *	Interp is marked for delete on next deallocation.
 *
 *----------------------------------------------------------------------
 */

static int
DebugInit(Tcl_Interp *interp, char *host, char *port, char *procs)
{
	AdpData *adPtr;
	Tcl_DString ds;
	int code;
	
	code = TCL_OK;
	adPtr = NsAdpGetData();
	if (adPtr->debugInit != NS_FALSE) {
		Tcl_DStringInit(&ds);
		Tcl_DStringAppendElement(&ds, debugInitProc);
		Tcl_DStringAppendElement(&ds, procs ? procs : "");
		Tcl_DStringAppendElement(&ds, host ? host : "");
		Tcl_DStringAppendElement(&ds, port ? port : "");
		code = Tcl_Eval(interp, ds.string);

		Tcl_DStringFree(&ds);
		if (code != TCL_OK) {
			Ns_TclLogError(interp);
			return TCL_ERROR;
		}
		if (Tcl_LinkVar(interp, "ns_adp_output",
			(char *) &adPtr->output.string,
			TCL_LINK_STRING | TCL_LINK_READ_ONLY) != TCL_OK) {
			Ns_TclLogError(interp);
		}
		adPtr->debugInit = 1;
		adPtr->debugLevel = 1;
	}
	
	return code;
}


/*
 *----------------------------------------------------------------------
 *
 * ParsePage --
 *
 *	Figure out which parser to use and then parse a page, putting 
 *	the result into the dstring.  
 *
 * Results:
 *	None. 
 *
 * Side effects:
 *	Appends to 'out'. If filename is null, it will use the default
 *	parser.
 *
 *----------------------------------------------------------------------
 */

static void
Parse(char *filename, Ns_DString *out, char *pagein)
{
	char			 *ext;
	Tcl_HashEntry	 *hePtr;
	Ns_AdpParserProc *parserProc;
	
	if (filename == NULL) {
		filename = "";
	}
	
	ext = strrchr(filename, '.');
	
	if (ext == NULL) {
		ext = "";
	}
	
	hePtr = Tcl_FindHashEntry(&extensionsTable, ext);
	if (hePtr == NULL) {
		parserProc = defParserProc;
	} else {
		parserProc = (Ns_AdpParserProc *) Tcl_GetHashValue(hePtr);
	}
	(*parserProc)(out, pagein);
}


/*
 *----------------------------------------------------------------------
 *
 * FreePage --
 *
 *   Callback function that is called after a time interval to
 *	free memory.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	None.
 *
 *----------------------------------------------------------------------
 */

static void
FreePage(Page *pagePtr)
{
	Ns_CacheFree(nsAdpCachePtr, pagePtr);
}
