/*
 * The contents of this file are subject to the AOLserver Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * http://aolserver.lcs.mit.edu/.
 *
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
 * the License for the specific language governing rights and limitations
 * under the License.
 *
 * The Original Code is AOLserver Code and related documentation
 * distributed by AOL.
 * 
 * The Initial Developer of the Original Code is America Online,
 * Inc. Portions created by AOL are Copyright (C) 1999 America Online,
 * Inc. All Rights Reserved.
 *
 * Alternatively, the contents of this file may be used under the terms
 * of the GNU General Public License (the "GPL"), in which case the
 * provisions of GPL are applicable instead of those above.	If you wish
 * to allow use of your version of this file only under the terms of the
 * GPL and not to allow others to use your version of this file under the
 * License, indicate your decision by deleting the provisions above and
 * replace them with the notice and other provisions required by the GPL.
 * If you do not delete the provisions above, a recipient may use your
 * version of this file under either the License or the GPL.
 */

/* 
 * cache.c --
 *
 *	Routines for a simple cache used by fastpath and Adp.
 *
 *	Based on cache.c in AOLserver 3.0 beta 3.	
 *
 */

#include "ns.h"
#include "adpAS21.h"

struct Cache;

/*
 * An Entry is a node in a linked list as well as being a
 * hash table entry. The linked list is there to keep track of
 * usage for the purposes of cache pruning.
 */

typedef struct Entry {
	struct Entry *nextPtr;
	struct Entry *prevPtr;
	struct Cache *cachePtr;
	Tcl_HashEntry *hPtr;
	time_t expires;
	size_t size;
	void *value;
} Entry;

/*
 * The following structure defines a cache
 */

typedef struct Cache {
	char	name[32];
	Entry *firstEntryPtr;
	Entry *lastEntryPtr;
	int keys;
	time_t timeout;
	size_t maxSize;
	size_t currentSize;
	Ns_Callback *freeProc;
	Ns_Pool *pool;
	Ns_Mutex lock;
	Ns_Cond cond;
	unsigned int nhit;
	unsigned int nmiss;
	unsigned int nflush;
	Tcl_HashTable entriesTable;
} Cache;


/*
 * Local functions defined in this file
 */

static Ns_TclInterpInitProc EnableCmds;
static Ns_Cache * CacheCreate(char *name, int keys, time_t timeout,
							  size_t maxSize, Ns_Callback *freeProc);
static int GetCache(Tcl_Interp *interp, char *name, Cache **cachePtrPtr);
static void PurgeProc(void *arg);
static void Delink(Entry *ePtr);
static void Push(Entry *ePtr);

/*
 * Static variables defined in this file
 */

static Tcl_HashTable cachesTable;	/* a hash table of hash tables */


/*
 *==========================================================================
 * API functions
 *==========================================================================
 */

/*
 *----------------------------------------------------------------------
 *
 * Ns_CacheCreate --
 *
 *	Calls CacheCreate with an unlimited max size.
 *
 * Results:
 *	See CacheCreate()
 *
 * Side effects:
 *	See CacheCreate()
 *
 *----------------------------------------------------------------------
 */

Ns_Cache *
Ns_CacheCreate(char *name, int keys, time_t timeout, Ns_Callback *freeProc)
{
	return CacheCreate(name, keys, timeout, 0, freeProc);
}

/*
 *----------------------------------------------------------------------
 *
 * Ns_CacheCreateSz --
 *
 *	See CacheCreate()
 *
 * Results:
 *	See CacheCreate()
 *
 * Side effects:
 *	See CacheCreate()
 *
 *----------------------------------------------------------------------
 */

Ns_Cache *
Ns_CacheCreateSz(char *name, int keys, size_t maxSize, Ns_Callback *freeProc)
{
	return CacheCreate(name, keys, -1, maxSize, freeProc);
}

/*
 *----------------------------------------------------------------------
 *
 * Ns_CacheFind --
 *
 *	Find a cache by name.
 *
 * Results:
 *	A pointer to an Ns_Cache or NULL 
 *
 * Side effects:
 *	None. 
 *
 *----------------------------------------------------------------------
 */

Ns_Cache *
Ns_CacheFind(char *name)
{
	Tcl_HashEntry *hPtr;
	
	hPtr = Tcl_FindHashEntry(&cachesTable, name);
	if (hPtr != NULL) {
		return (Ns_Cache *) Tcl_GetHashValue(hPtr);
	}
	
	return NULL;
}

/*
 *----------------------------------------------------------------------
 *
 * Ns_CacheMalloc --
 *
 *	Allocate memory from a cache-local pool. 
 *
 * Results:
 *	A pointer to new memory.
 *
 * Side effects:
 *	See Ns_PoolAlloc 
 *
 *----------------------------------------------------------------------
 */

void *
Ns_CacheMalloc(Ns_Cache *cache, size_t len)
{
    Cache *cachePtr = (Cache *) cache;

    return Ns_PoolAlloc(cachePtr->pool, len);
}


/*
 *----------------------------------------------------------------------
 *
 * Ns_CacheFree --
 *
 *	Frees memory allocated from Ns_CacheMalloc 
 *
 * Results:
 *	None. 
 *
 * Side effects:
 *	See Ns_PoolFree 
 *
 *----------------------------------------------------------------------
 */

void
Ns_CacheFree(Ns_Cache *cache, void *bytes)
{
	Cache *cachePtr = (Cache *) cache;
	
	Ns_PoolFree(cachePtr->pool, bytes);
}

/*
*----------------------------------------------------------------------
*
* Ns_CacheFindEntry --
*
*	Find a cache entry 
*
* Results:
*	A pointer to an Ns_Entry cache entry 
*
* Side effects:
*	The cache entry will move to the top of the LRU list. 
*
*----------------------------------------------------------------------
*/

Ns_Entry *
Ns_CacheFindEntry(Ns_Cache *cache, char *key)
{
	Cache *cachePtr = (Cache *) cache;
	Tcl_HashEntry *hPtr;
	Entry *ePtr;
	
	hPtr = Tcl_FindHashEntry(&cachePtr->entriesTable, key);
	if (hPtr == NULL) {
		++cachePtr->nmiss;
		return NULL;
	}
	
	++cachePtr->nhit;
	ePtr = Tcl_GetHashValue(hPtr);
	Delink(ePtr);
	Push(ePtr);
	
	return (Ns_Entry *) ePtr;
}

/*
 *----------------------------------------------------------------------
 *
 * Ns_CacheCreateEntry --
 *
 *	Create a new cache entry; this function emulates 
 *	Tcl_CreateHashEntry's interface 
 *
 * Results:
 *	A pointer to a new cache entry 
 *
 * Side effects:
 *	Memory will be allocated for the new cache entry and it will 
 *	be inserted into the cache. 
 *
 *----------------------------------------------------------------------
 */

Ns_Entry *
Ns_CacheCreateEntry(Ns_Cache *cache, char *key, int *newPtr)
{
	Cache *cachePtr = (Cache *) cache;
	Tcl_HashEntry *hPtr;
	Entry *ePtr;
	
	hPtr = Tcl_CreateHashEntry(&cachePtr->entriesTable, key, newPtr);
	if (*newPtr == 0) {
		ePtr = Tcl_GetHashValue(hPtr);
		Delink(ePtr);
		++cachePtr->nhit;
	} else {
		ePtr = Ns_CacheMalloc(cache, sizeof(Entry));
		ePtr->prevPtr = ePtr->nextPtr = NULL;
		ePtr->value = NULL;
		ePtr->size = 0;
		ePtr->hPtr = hPtr;
		ePtr->expires = 0;
		ePtr->cachePtr = cachePtr;
		Tcl_SetHashValue(hPtr, ePtr);
		++cachePtr->nmiss;
	}
	Push(ePtr);
	
	return (Ns_Entry *) ePtr;
}

/*
 *----------------------------------------------------------------------
 *
 * Ns_CacheName --
 *
 *	Gets the name of the cache 
 *
 * Results:
 *	A pointer to a null-terminated string 
 *
 * Side effects:
 *	None. 
 *
 *----------------------------------------------------------------------
 */

char *
Ns_CacheName(Ns_Entry *entry)
{
	Entry *ePtr = (Entry *) entry;
	
	return ePtr->cachePtr->name;
}

/*
 *----------------------------------------------------------------------
 *
 * Ns_CacheKey --
 *
 *	Gets the key of a cache entry 
 *
 * Results:
 *	A pointer to the key for the given entry 
 *
 * Side effects:
 *	None. 
 *
 *----------------------------------------------------------------------
 */

char *
Ns_CacheKey(Ns_Entry *entry)
{
	Entry *ePtr = (Entry *) entry;
	
	return Tcl_GetHashKey(&ePtr->cachePtr->entriesTable, ePtr->hPtr);
}

/*
 *----------------------------------------------------------------------
 *
 * Ns_CacheGetValue --
 *
 *	Get the value (contents) of a cache entry 
 *
 * Results:
 *	A pointer to the cache entry's contents 
 *
 * Side effects:
 *	None. 
 *
 *----------------------------------------------------------------------
 */

void *
Ns_CacheGetValue(Ns_Entry *entry)
{
	Entry *ePtr = (Entry *) entry;
	
	return ePtr->value;
}

/*
 *----------------------------------------------------------------------
 *
 * Ns_CacheSetValue --
 *
 *	Set the value of a cache entry 
 *
 * Results:
 *	None. 
 *
 * Side effects:
 *	See Ns_CacheSetValueSz 
 *
 *----------------------------------------------------------------------
 */

void
Ns_CacheSetValue(Ns_Entry *entry, void *value)
{
	Ns_CacheSetValueSz(entry, value, 0);
}

/*
 *----------------------------------------------------------------------
 *
 * Ns_CacheSetValueSz --
 *
 *	Free the cache entry's previous contents, set it to the new 
 *	contents, increase the size of the cache, and prune until 
 *	it's back under the maximum size. 
 *
 * Results:
 *	None. 
 *
 * Side effects:
 *	Cache pruning and freeing of old contents may occur. 
 *
 *----------------------------------------------------------------------
 */

void
Ns_CacheSetValueSz(Ns_Entry *entry, void *value, size_t size)
{
	Entry *ePtr = (Entry *) entry;
	Cache *cachePtr = ePtr->cachePtr;
	
	Ns_CacheUnsetValue(entry);
	ePtr->value = value;
	ePtr->size = size;
	cachePtr->currentSize += size;
	if (ePtr->cachePtr->maxSize > 0) {
		while (cachePtr->currentSize > cachePtr->maxSize &&
			cachePtr->lastEntryPtr != ePtr) {
			
			Ns_CacheFlushEntry((Ns_Entry *) cachePtr->lastEntryPtr);
		}
	}
}

/*
 *----------------------------------------------------------------------
 *
 * Ns_CacheUnsetValue --
 *
 *	Reset the value of an entry to NULL, calling the free proc for
 *	any previous entry an updating the cache size.
 *
 * Results:
 *		None.
 *
 * Side effects:
 *	None.
 *
 *----------------------------------------------------------------------
 */

void
Ns_CacheUnsetValue(Ns_Entry *entry)
{
	Entry *ePtr = (Entry *) entry;
	Cache *cachePtr;
	
	if (ePtr->value != NULL) {
		cachePtr = ePtr->cachePtr;
		cachePtr->currentSize -= ePtr->size;
		if (cachePtr->freeProc == NS_CACHE_FREE) {
			Ns_CacheFree((Ns_Cache *) cachePtr, ePtr->value);
		} else if (cachePtr->freeProc != NULL) {
			(*cachePtr->freeProc)(ePtr->value);
		}
		ePtr->size = 0;
		ePtr->value = NULL;
	}
}

/*
 *----------------------------------------------------------------------
 *
 * Ns_CacheDeleteEntry --
 *
 *	Delete an entry from the cache table.
 *
 * Results:
 *		None.
 *
 * Side effects:
 *	None.
 *
 *----------------------------------------------------------------------
 */

void
Ns_CacheDeleteEntry(Ns_Entry *entry)
{
	Entry *ePtr = (Entry *) entry;
	
	Delink(ePtr);
	Tcl_DeleteHashEntry(ePtr->hPtr);
	Ns_CacheFree((Ns_Cache *) ePtr->cachePtr, ePtr);
}

/*
 *----------------------------------------------------------------------
 *
 * Ns_CacheFlushEntry --
 *
 *	Delete an entry from the cache table after first unsetting
 *		the current entry value (if any).
 *
 * Results:
 *		None.
 *
 * Side effects:
 *	None.
 *
 *----------------------------------------------------------------------
 */

void
Ns_CacheFlushEntry(Ns_Entry *entry)
{
	Entry *ePtr = (Entry *) entry;
	
	++ePtr->cachePtr->nflush;
	Ns_CacheUnsetValue(entry);
	Ns_CacheDeleteEntry(entry);
}

/*
 *----------------------------------------------------------------------
 *
 * Ns_CacheFirstEntry --
 *
 *	Return a pointer to the first entry in the cache (in no 
 *	particular order) 
 *
 * Results:
 *	A pointer to said entry. 
 *
 * Side effects:
 *	None. 
 *
 *----------------------------------------------------------------------
 */

Ns_Entry *
Ns_CacheFirstEntry(Ns_Cache *cache, Ns_CacheSearch *search)
{
	Tcl_HashSearch *sPtr = (Tcl_HashSearch *) search;
	Cache *cachePtr = (Cache *) cache;
	Tcl_HashEntry *hPtr;
	
	hPtr = Tcl_FirstHashEntry(&cachePtr->entriesTable, sPtr);
	if (hPtr == NULL) {
		return NULL;
	}
	
	return (Ns_Entry *) Tcl_GetHashValue(hPtr);
}

/*
 *----------------------------------------------------------------------
 *
 * Ns_CacheNextEntry --
 *
 *	When used in conjunction with Ns_CacheFirstEntry, one may 
 *	walk through the whole cache. 
 *
 * Results:
 *	NULL or a pointer to an entry 
 * 
 * Side effects:
 *	None. 
 *
 *----------------------------------------------------------------------
 */

Ns_Entry *
Ns_CacheNextEntry(Ns_CacheSearch *search)
{
	Tcl_HashSearch *sPtr = (Tcl_HashSearch *) search;
	Tcl_HashEntry *hPtr;
	
	hPtr = Tcl_NextHashEntry(sPtr);
	if (hPtr == NULL) {
		return NULL;
	}
	
	return (Ns_Entry *) Tcl_GetHashValue(hPtr);
}

/*
 *----------------------------------------------------------------------
 *
 * Ns_CacheFlush --
 *
 * 	Flush every entry from a cache. 
 *
 * Results:
 *		None.
 *
 * Side effects:
 *	None.
 *
 *----------------------------------------------------------------------
 */

void
Ns_CacheFlush(Ns_Cache *cache)
{
	Ns_CacheSearch search;
	Ns_Entry *entry;
	
	entry = Ns_CacheFirstEntry(cache, &search);
	while (entry != NULL) {
		Ns_CacheFlushEntry(entry);
		entry = Ns_CacheNextEntry(&search);
	}
}

/*
 *----------------------------------------------------------------------
 *
 * Ns_CacheLock --
 *
 *	Lock the cache
 *
 * Results:
 *		None.
 *
 * Side effects:
 *	Mutex locked.
 *
 *----------------------------------------------------------------------
 */

void
Ns_CacheLock(Ns_Cache *cache)
{
	Cache *cachePtr = (Cache *) cache;
	
	Ns_LockMutex(&cachePtr->lock);
}

/*
 *----------------------------------------------------------------------
 *
 * Ns_CacheUnlock --
 *
 *	Unlock the cache entry
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	Mutex unlocked.
 *
 *----------------------------------------------------------------------
 */

void
Ns_CacheUnlock(Ns_Cache *cache)
{
	Cache *cachePtr = (Cache *) cache;
	
	Ns_UnlockMutex(&cachePtr->lock);
}

//
// Cache functions removed. These function rely on Ns_Cond* functions
// which aren't in 2.1 and would be very difficult to implement. Since
// they're not needed by ADPs we've removed them.
/*
 *----------------------------------------------------------------------
 *
 * Ns_CacheTimedGetValue --
 *
 *	Wait for an entry's value to be set to non-null by some other
 *	thread up to the given timeout or until an optional condition
 *	integer becomes zero.  Note that the cache and key are given
 *	instead of the entry because you cannot rely on an entry to
 *	remain valid during the Ns_CondTimedWait.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	Thread is suspended until entry is available or timeout.
 *
 *----------------------------------------------------------------------
 */
//void *
//Ns_CacheTimedGetValue(Ns_Cache *cache, char *key, Ns_Time *timePtr, int *okPtr)
//{
//	Cache *cachePtr = (Cache *) cache;
//	Entry *ePtr;
//	int ok;
//	
//	/*
//	* If okPtr is NULL, set it to a local variable which is
//	* always 1.
//	*/
//	
//	if (okPtr == NULL) {
//		ok = 1;
//		okPtr = &ok;
//	}
//	
//	/*
// 	 * Wait for the entries value to become non-NULL up to the given
//	 * timeout as long as the value at *okPtr is still non-zero.
//	 */
//	
//	while (*okPtr &&
//		(ePtr = (Entry *) Ns_CacheFindEntry(cache, key)) != NULL &&
//		ePtr->value == NULL &&
//		Ns_CondTimedWait(&cachePtr->cond, &cachePtr->lock, timePtr) == NS_OK) {
//	while (*okPtr &&
//		(ePtr = (Entry *) Ns_CacheFindEntry(cache, key)) != NULL &&
//		ePtr->value == NULL) {
//
//	
//		/* NULL STATEMENT */ ;
//	}
//	
//	return (ePtr ? ePtr->value : NULL);
//}

/*
 *----------------------------------------------------------------------
 *
 * Ns_CacheTimedWait --
 *
 *	Wait for the cache's condition variable to be
 *	signaled or the given absolute timeout if timePtr is not NULL.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	Thread is suspended until condition is signaled or timeout.
 *
 *----------------------------------------------------------------------
 */

//int
//Ns_CacheTimedWait(Ns_Cache *cache, Ns_Time *timePtr)
//{
//	Cache *cachePtr = (Cache *) cache;
//	
//	return Ns_CondTimedWait(&cachePtr->cond, &cachePtr->lock, timePtr);
//}

/*
 *----------------------------------------------------------------------
 *
 * Ns_CacheWait --
 *
 *	Wait indefinitely for the cache's condition variable to be
 *		signaled.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	Thread is suspended until condition is signaled.
 * 
 *----------------------------------------------------------------------
 */

//void
//Ns_CacheWait(Ns_Cache *cache)
//{
//	Ns_CacheTimedWait(cache, NULL);
//}

/*
 *----------------------------------------------------------------------
 *
 * Ns_CacheSignal --
 *
 *	Signal the cache's condition variable, waking the first waiting
 *	thread (if any).
 *
 *	NOTE:  Be sure you don't really want to wake all threads with
 *	Ns_CacheBroadcast.
 *
 * Results: 
 *	None.
 *
 * Side effects:
 *	A single thread may resume.
 *
 *----------------------------------------------------------------------
 */

//void
//Ns_CacheSignal(Ns_Cache *cache)
//{
//	Cache *cachePtr = (Cache *) cache;
//	
//	Ns_CondSignal(&cachePtr->cond);
//}


/*
 *----------------------------------------------------------------------
 *
 * Ns_CacheBroadcast --
 *
 *	Broadcast the cache's condition variable, waking all waiting
 *	threads (if any).
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	Threads may resume.
 *
 *----------------------------------------------------------------------
 */

//void
//Ns_CacheBroadcast(Ns_Cache *cache)
//{
//	Cache *cachePtr = (Cache *) cache;
//	
//	Ns_CondBroadcast(&cachePtr->cond);
//}


/* 
 *==========================================================================
 * Exported functions
 *==========================================================================
 */

/*
 *----------------------------------------------------------------------
 *
 * NsCacheInit --
 *
 *	Initailize the cache
 *
 * Results:
 *	NS_OK.
 *
 * Side effects:
 *	Allocate some memory, create some Tcl commands
 *
 *----------------------------------------------------------------------
 */

void
NsCacheInit(char *server)
{

	Tcl_InitHashTable(&cachesTable, TCL_STRING_KEYS);
	Ns_TclInitInterps(server, EnableCmds, NULL);
}

/*
 *----------------------------------------------------------------------
 *
 * NsTclNamesCmd --
 *
 *	Spit back a list of cache names
 *
 * Results:
 *	Tcl result.
 *
 * Side effects:
 *	A list of cache names will be appended to the interp->result.
 *
 *----------------------------------------------------------------------
 */

int
NsTclNamesCmd(ClientData dummy, Tcl_Interp *interp, int argc, char **argv)
{
	Tcl_HashEntry *hPtr;
	Tcl_HashSearch search;
	
	if (argc != 1) {
		Tcl_AppendResult(interp, "wrong # args: should be \"",
			argv[0], "\"", NULL);
		return TCL_ERROR;
	}
	
	hPtr = Tcl_FirstHashEntry(&cachesTable, &search);
	while (hPtr != NULL) {
		Tcl_AppendElement(interp, Tcl_GetHashKey(&cachesTable, hPtr));
		hPtr = Tcl_NextHashEntry(&search);
	}
	
	return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * NsTclStatsCmds --
 *
 *	Returns stats on a cache
 *
 * Results:
 *	Tcl result.
 *
 * Side effects:
 *	Results will be appended to interp.
 *
 *----------------------------------------------------------------------
 */

int
NsTclStatsCmd(ClientData dummy, Tcl_Interp *interp, int argc, char **argv)
{
	Cache *cachePtr;
	char buf[20];
	int entries, flushed, hits, misses, total, hitrate;
	
	if (argc != 2 && argc != 3) {
		Tcl_AppendResult(interp, "wrong # args: should be \"",
			argv[0], " cache ?arrayVar?\"", NULL);
		return TCL_ERROR;
	}
	if (GetCache(interp, argv[1], &cachePtr) != TCL_OK) {
		return TCL_ERROR;
	}
	
	Ns_LockMutex(&cachePtr->lock);
	entries = cachePtr->entriesTable.numEntries;
	flushed = cachePtr->nflush;
	hits = cachePtr->nhit;
	misses = cachePtr->nmiss;
	total = cachePtr->nhit + cachePtr->nmiss;
	hitrate = (total ? (cachePtr->nhit * 100) / total : 0);
	Ns_UnlockMutex(&cachePtr->lock);
	
	if (argc == 2) {
		sprintf(interp->result,
			"entries: %d  flushed: %d  hits: %d  misses: %d  hitrate: %d",
			entries, flushed, hits, misses, hitrate);
	} else {
		sprintf(buf, "%d", entries);
		if (Tcl_SetVar2(interp, argv[2], "entries", buf,
			TCL_LEAVE_ERR_MSG) == NULL) {
			return TCL_ERROR;
		}
		sprintf(buf, "%d", flushed);
		if (Tcl_SetVar2(interp, argv[2], "flushed", buf,
			TCL_LEAVE_ERR_MSG) == NULL) {
			return TCL_ERROR;
		}
		sprintf(buf, "%d", hits);
		if (Tcl_SetVar2(interp, argv[2], "hits", buf,
			TCL_LEAVE_ERR_MSG) == NULL) {
			return TCL_ERROR;
		}
		sprintf(buf, "%d", misses);
		if (Tcl_SetVar2(interp, argv[2], "misses", buf,
			TCL_LEAVE_ERR_MSG) == NULL) {
			return TCL_ERROR;
		}
		sprintf(buf, "%d", hitrate);
		if (Tcl_SetVar2(interp, argv[2], "hitrate", buf,
			TCL_LEAVE_ERR_MSG) == NULL) {
			return TCL_ERROR;
		}
	}
	
	return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * NsTclFlushCmd --
 *
 *	Wipe out a cache entry
 *
 * Results:
 *	TCL result.
 *
 * Side effects:
 *	A cache will be cleared.
 *
 *----------------------------------------------------------------------
 */

int
NsTclFlushCmd(ClientData dummy, Tcl_Interp *interp, int argc, char **argv)
{
	Ns_Cache *cache;
	Ns_Entry *entry;
	
	if (argc != 3) {
		Tcl_AppendResult(interp, "wrong # args: should be \"",
			argv[0], " cache key?\"", NULL);
		return TCL_ERROR;
	}
	if (GetCache(interp, argv[1], (Cache **) &cache) != TCL_OK) {
		return TCL_ERROR;
	}
	
	Ns_CacheLock(cache);
	entry = Ns_CacheFindEntry(cache, argv[2]);
	if (entry != NULL) {
		Ns_CacheFlushEntry(entry);
	}
	Ns_CacheUnlock(cache);
	
	return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * NsTclSizeCmd --
 *
 *	Returns current size of a cache.
 *
 * Results:
 *	Tcl result.
 *
 * Side effects:
 *	Results will be appended to interp.
 *
 *----------------------------------------------------------------------
 */

int
NsTclSizeCmd(ClientData dummy, Tcl_Interp *interp, int argc, char **argv)
{
	Cache *cachePtr;
	size_t maxSize, currentSize;
	
	if (argc != 2) {
		Tcl_AppendResult(interp, "wrong # args: should be \"",
			argv[0], " cache\"", NULL);
		return TCL_ERROR;
	}
	if (GetCache(interp, argv[1], &cachePtr) != TCL_OK) {
		return TCL_ERROR;
	}
	Ns_LockMutex(&cachePtr->lock);
	maxSize = cachePtr->maxSize;
	currentSize = cachePtr->currentSize;
	Ns_UnlockMutex(&cachePtr->lock);
	sprintf(interp->result, "%ld %ld", (long) maxSize, (long) currentSize);
	
	return TCL_OK;
}

/*
 *==========================================================================
 * Static functions
 *==========================================================================
 */

/*
 *----------------------------------------------------------------------
 *
 * EnableCmds --
 *
 *	Create the cache Tcl commands.
 *
 * Results:
 *	NS_OK.
 *
 * Side effects:
 *	Tcl commands are created in the parent Tcl interp.
 *
 *----------------------------------------------------------------------
 */

static int
EnableCmds(Tcl_Interp *interp, void *ignored)
{
	/*
	 * Tcl commands were added in tclcmds.c
	 */
	
	return NS_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * CacheCreate -- 
 *
 *	Create a new time or size based cache.
 *
 * Results:
 *	A pointer to the new cache.
 *
 * Side effects:
 *	Hash table is allocated, new pool created. A scheduled proc is
 *	created that will run every timeout seconds to flush the cache
 *	if timeout is greater than zero.
 * 
 *----------------------------------------------------------------------
 */

static Ns_Cache *
CacheCreate(char *name, int keys, time_t timeout, size_t maxSize,
			Ns_Callback *freeProc)
{
	Cache *cachePtr;
	Tcl_HashEntry *hPtr;
	int new;
	
	cachePtr = ns_malloc(sizeof(Cache));
	cachePtr->firstEntryPtr = cachePtr->lastEntryPtr = NULL;
	cachePtr->freeProc = freeProc;
	cachePtr->timeout = timeout;
	cachePtr->maxSize = maxSize;
	cachePtr->currentSize = 0;
	cachePtr->keys = keys;
	strcpy(cachePtr->name, name);
	cachePtr->pool = Ns_PoolCreate(cachePtr->name);
	cachePtr->nflush = cachePtr->nhit = cachePtr->nmiss = 0;
	Ns_InitializeMutex(&cachePtr->lock);

	/* 
	 * Since we've removed the functions which use Ns_Cond* this line is no 
	 * longer needed.
	 */
	//Ns_CondInit(&cachePtr->cond);
	
	Tcl_InitHashTable(&cachePtr->entriesTable, keys);
	if (timeout > 0) {
		Ns_ScheduleProc(PurgeProc, cachePtr, 0, timeout);
	}
	hPtr = Tcl_CreateHashEntry(&cachesTable, name, &new);
	Tcl_SetHashValue(hPtr, cachePtr);
	
	return (Ns_Cache *) cachePtr;
}

/*
 *----------------------------------------------------------------------
 *
 * Delink --
 *
 *	Remove a cache entry from the linked list of entries; this
 *	is used for maintaining the LRU list as well as removing entries
 *	that are still in use
 * 
 * Results:
 *	None.
 *
 * Side effects:
 *	The linked list will be changed.
 *
 *----------------------------------------------------------------------
 */

static void
Delink(Entry *ePtr)
{
	if (ePtr->prevPtr != NULL) {
		ePtr->prevPtr->nextPtr = ePtr->nextPtr;
	} else {
		ePtr->cachePtr->firstEntryPtr = ePtr->nextPtr;
	}
	if (ePtr->nextPtr != NULL) {
		ePtr->nextPtr->prevPtr = ePtr->prevPtr;
	} else {
		ePtr->cachePtr->lastEntryPtr = ePtr->prevPtr;
	}
	ePtr->prevPtr = ePtr->nextPtr = NULL;
}

/*
 *----------------------------------------------------------------------
 *
 * Push --
 *
 *	Stick an entry at the top of the linked list of entries, making
 *	it the Most Recently Used
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	The linked list will be changed and the expires time will be updated.
 *
 *----------------------------------------------------------------------
 */

static void
Push(Entry *ePtr)
{
	ePtr->expires = time(NULL) + ePtr->cachePtr->timeout;
	if (ePtr->cachePtr->firstEntryPtr != NULL) {
		ePtr->cachePtr->firstEntryPtr->prevPtr = ePtr;
	}
	ePtr->prevPtr = NULL;
	ePtr->nextPtr = ePtr->cachePtr->firstEntryPtr;
	ePtr->cachePtr->firstEntryPtr = ePtr;
	if (ePtr->cachePtr->lastEntryPtr == NULL) {
		ePtr->cachePtr->lastEntryPtr = ePtr;
	}
}

/*
 *----------------------------------------------------------------------
 *
 * GetCache --
 *
 *	Get a cache by name
 *
 * Results:
 *	Tcl result.
 * 
 * Side effects:
 *	*cachePtrPtr will contain a pointer to the cache if TCL_OK is
 *		returned.
 *
 *----------------------------------------------------------------------
 */

static int
GetCache(Tcl_Interp *interp, char *name, Cache **cachePtrPtr)
{
	*cachePtrPtr = (Cache *) Ns_CacheFind(name);
	if (*cachePtrPtr == NULL) {
		Tcl_AppendResult(interp, "no such cache: ", name, NULL);
		return TCL_ERROR;
	}
	
	return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 * 
 * PurgeProc --
 *
 *	Call free procs for all entries that have expired and
 *		delete them.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	Expired entries will be removed.
 *
 *----------------------------------------------------------------------
 */

static void
PurgeProc(void *arg)
{
	Entry *ePtr;
	Cache *cachePtr = (Cache *) arg;
	time_t now;
	
	time(&now);
	Ns_LockMutex(&cachePtr->lock);
	while ((ePtr = cachePtr->lastEntryPtr) != NULL &&
		ePtr->expires < now) {
		
		Ns_CacheFlushEntry((Ns_Entry *) ePtr);
	}
	Ns_UnlockMutex(&cachePtr->lock);
}