/*
 * The contents of this file are subject to the AOLserver Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * http://aolserver.lcs.mit.edu/.
 *
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
 * the License for the specific language governing rights and limitations
 * under the License.
 *
 * The Original Code is the ASP-style ADP module.
 * 
 * The Initial Developer of the Original Code is AM Computers.
 * Portions created by AM are Copyright (C) 1999, 2000 AM Computers.
 *
 * Alternatively, the contents of this file may be used under the terms
 * of the GNU General Public License (the "GPL"), in which case the
 * provisions of GPL are applicable instead of those above.  If you wish
 * to allow use of your version of this file only under the terms of the
 * GPL and not to allow others to use your version of this file under the
 * License, indicate your decision by deleting the provisions above and
 * replace them with the notice and other provisions required by the GPL.
 * If you do not delete the provisions above, a recipient may use your
 * version of this file under either the License or the GPL.
 */

/*
 * nsAspStyleAdps.c --
 *
 *	Registers the aspStyle ADP parser which supports partial Tcl 
 *	statements (ASP-style). Based on idea by Rob Schoening.  See 
 *	aolserver.am.net/srccontrib/adp/index.htm for latest version and 
 *	installation instructions.
 *
 *	Modified to work with AS 2.1 ADP port as a "compiled in" parser.
 */

#include "ns.h"
#include "adpAS21.h"

/*
 * Local functions defined in this file.
 */

static      Ns_TclInterpInitProc AddCmds;
static void AspStyleParsePage(Ns_DString *dsPtr, char *page);
static void AspStyleTextChunk(Ns_DString *dsPtr, char *text);

/* 
 *==========================================================================
 * Exported functions
 *==========================================================================
 */

/*
 *----------------------------------------------------------------------
 * NsAspStyleAdpsInit
 *
 *	Registers aspStyle ADP parser and adds _ns_asp_puts
 *	Tcl command
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	Registers aspStyle parser and adds _ns_asp_puts.
 *
 *----------------------------------------------------------------------
 */

void
NsAspStyleAdpsInit(char *hServer, char *path)
{
    Ns_AdpRegisterParser("aspStyle", *AspStyleParsePage); 
    Ns_TclInitInterps(hServer, AddCmds, NULL);

}


/*
 *==========================================================================
 * Static functions
 *==========================================================================
 */

/*
 *----------------------------------------------------------------------
 *
 * AddCmds --
 *
 *	Adds _ns_asp_puts Tcl procedure which is a wrapper to ns_adp_puts.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	_ns_asp_puts added.
 *
 *----------------------------------------------------------------------
 */

static int
AddCmds(Tcl_Interp *interp, void *ctx)
{
    Tcl_GlobalEval(interp, "proc _ns_asp_puts args {\n\tforeach textChunk $args {\n\t\tns_adp_puts -nonewline $textChunk\n\t}\n}");
    
    return NS_OK;
}


/*
 *----------------------------------------------------------------------
 *
 * AspStyleParsePage --
 *
 *	Non-fancy ADP parser allowing for partial Tcl statements in ADP
 *	chunks. We acomplish this by making everything one gaint
 *	script chunk. We add a _ns_asp_puts around what use to be a 
 *	text chunk. This code is based on ParsePage in nsd/asp.c. It's a 
 *	little tricky so study it carefully.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	Chunks are parsed into given Ns_DString.
 *
 *----------------------------------------------------------------------
 */

static void
AspStyleParsePage(Ns_DString *dsPtr, char *page)
{
    register char *s, *e;
    register char *t = page;
    
    Ns_DStringNAppend(dsPtr, "s", 1);
    while ((s = strstr(t, "<%")) != NULL && 
	   (e = strstr(s, "%>")) != NULL) {
	
	*s = '\0';
	AspStyleTextChunk(dsPtr, t);
	s += 2;
	
	if (*s == '=') {
	    Ns_DStringAppend(dsPtr, "\n_ns_asp_puts ");
	    ++s;
	}
	*e = '\0';
	Ns_DStringNAppend(dsPtr, s, e-s);
	t = e+2;
    }
    AspStyleTextChunk(dsPtr, t);
    Ns_DStringNAppend(dsPtr, "\0", 1);	 
}

/*
 *----------------------------------------------------------------------
 *
 * AspStyleTextChunk --
 *
 *	A healper to AspStyleParsePage. Builds a "text chunk" for the
 *	the ASP-Style ADP parser. 
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	Text chunk is parsed into given Ns_DString.
 *
 *----------------------------------------------------------------------
 */

static void
AspStyleTextChunk(Ns_DString *dsPtr, char *text)
{
    register char *b = NULL, *o = NULL;
  
    Ns_DStringAppend(dsPtr, "\n_ns_asp_puts {");

    /*
     * Because a text chunk is within a _ns_asp_puts {} we must 
     * "escape" curly braces. We could escape them by putting a \
     * in front of them, but since the brace is in braces 
     * Tcl's subsition rules will also include the \ in the 
     * string. Instead we must pass them as an individual paramater 
     * to _ns_asp_puts and enclose the brace in quotes.
     */
    
    /* The expression in the while loop and the if below are 
     * intended to make you appreciate Tcl. Basically we're justing
     * making sure that b is the first brace (either opening or closing)
     * in text. 
     */

    // Bug in this while condition found and fixed by Ryder Rishel; Sat, 4 Dec 1999
    while ((((b = strchr(text, '}')) != NULL) 
	    && ((b < (o = strchr(text, '{'))) || (o == NULL)))
	   || ((b = strchr(text, '{')) != NULL && (o = b))) {
	
	if ((b > o) && (b != NULL) && (o != NULL)) {
	    b = o;
	}
	
	/*
	 * Write everything up to the brace
	 */
	Ns_DStringNAppend(dsPtr, text, b-text);
	
	Ns_DStringAppend(dsPtr, "} \"");
	Ns_DStringNAppend(dsPtr, "\\", 1);
	Ns_DStringNAppend(dsPtr, b, 1);
	Ns_DStringAppend(dsPtr, "\" {");
	
	/*
	 * Move our postion in text chunk to right after brace
	 */
	text = (b+1);
    }
    Ns_DStringNAppend(dsPtr, text, strlen(text));
    
    Ns_DStringAppend(dsPtr, "}\n");
}
